/* *********************************************************** */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                 */
/*                                                             */
/*  Cholesky-Zerlegung einer Bandmatrix                        */
/*  ===================================                        */
/*                                                             */
/*  Fuer die symmetrische, positiv definite Matrix A           */
/*  wird die Dreieckszerlegung nach Cholesky                   */
/*                                                             */
/*                           T                                 */
/*                      A = R  R                               */
/*                                                             */
/*  ausgefuehrt. Von A sind ab a_p nur die wesentlichen Ele-   */
/*  mente zeilenweise in einem eindimensionalen Feld gespei-   */
/*  chert (pro Zeile ib Elemente, ib ist die Bandweite).       */
/*  Die letzten ib - 1 Zeilen muessen durch Null-Elemente      */
/*  auf jeweils ib Elemente ergaenzt werden (Beschreibung      */
/*  des Loesungsverfahrens z. B. in  "Dankert: Numerische      */
/*  Methoden der Mechanik", Springer-Verlag 1977).             */
/*                                                             */
/*  Beispiel:                                                  */
/*                                                             */
/*  Symmetrische Bandmatrix     Elemente auf dem Feld a_p      */
/*                                                             */
/*      4  2  0  0  0  0          4  2  0                      */
/*      2  6  2  3  0  0             6  2  3                   */
/*      0  2  7  1  4  0                7  1  4                */
/*      0  0  1  5  9  9     --->          5  9  9             */
/*      0  3  4  9  1  2                      1  2   0         */
/*      0  0  0  9  2  2                         2   0  0      */
/*                                                             */
/*  Diese Matrix hat die Bandweite ib = 3. Die Elemente        */
/*  muessen in folgender Reihenfolge auf dem eindimensiona-    */
/*  len Feld A stehen:                                         */
/*                                                             */
/*    4  2  0  6  2  3  7  1  4  5  9  9  1  2  0  2  0  0     */
/*                                                             */
/*  Eingabe:   a_p   -  Bandmatrix, fuer die die Cholesky-     */
/*                      Zerlegung ausgefuehrt werden soll      */
/*                      (eindimensionales Feld mit n*ib        */
/*                      Elementen)                             */
/*             n     -  Zeilen- bzw. Spaltenanzahl der         */
/*                      Matrix A                               */
/*             ib    -  Bandweite der Matrix A                 */
/*             eps   -  Schranke fuer Abbruchkriterium (wenn   */
/*                      ein Hauptdiagonalelement um das eps-   */
/*                      fach kleiner wird als das kleinste     */
/*                      vorher entstandene Hauptdiagonal-      */
/*                      element, wird "nicht positiv definite  */
/*                      Matrix" vermutet, und die Rechnung     */
/*                      wird abgebrochen)                      */
/*                                                             */
/*  Ausgabe    a_p   -  Rechtsdreiecksmatrix R, Bandmatrix,    */
/*                      die auf die gleiche Art wie die        */
/*                      Matrix A zeilenweise gespeichert       */
/*                      ist (Eingabematrix wird ueber-         */
/*                      schrieben)                             */
/*       Return-Wert:   Indikator, der den Erfolg der          */
/*                      Berechnung anzeigt:                    */
/*                      = 1  -->  Matrix A ist positiv         */
/*                                definit, Berechnung von      */
/*                                R wurde ausgefuehrt          */
/*                      = 0  -->  Matrix A ist nicht posi-     */
/*                                tiv definit, Berechnung      */
/*                                wurde abgebrochen            */
/*                                                             */
/*  Autor: J. Dankert                                          */
/* *********************************************************** */

#include    <math.h>
#include    "femset_f.h"

int cholba_f (double *a_p , long n , long ib , double eps)
{
   long    ib1 , irue , i , ie , j , iee , kkk , kk , k , ia , iaa ;
   double  ss  , smin , s ;
   int     succ = 0 ;

   fillin_f (1 , 0L , 0L) ;
      
   ib1  = ib  - 1       ;
   irue = ib1 * ib1 - 1 ;

   if (*(a_p) <= 0.) goto NonPosDef ;

   *(a_p) = sqrt (*(a_p)) ;

   if (ib <= 1)
   {
     if (n < 2) goto PosDef ;
     for (i = 1 ; i < n ; i++)
     {
       s = *(a_p+i)              ;
       if (s <= 0.) goto NonPosDef ;
       *(a_p+i) = sqrt (s)       ;
       fillin_f (50 , i , 0L) ;
     }
     goto PosDef ;
   }


   for (i = 1 ; i < ib ; i++)
     *(a_p+i) /= *(a_p) ;

   fillin_f (50 , 1 , 0L) ;

   if (ib > 2)
   {
     ie = ib ;
     for (i = 2 ; i <= ib1 ; i++)
     {
       s = *(a_p+ie) ;
       for (j = i-1 ; j < ie ; j += ib1)
         s -= *(a_p+j) * *(a_p+j) ;

       if (s <= 0.) goto NonPosDef ;

       if (i == 2)
         smin = s ;
       else
       {
         if (s < smin * eps) goto NonPosDef ;
         if (s < smin)       smin = s ;
       }

       s   = sqrt (s) ;
       *(a_p+ie) = s  ;
       s   = 1. / s  ;
       iee = ie   + 1 ;
       kkk = i        ;

       for (j = i+1 ; j <= ie ; j++ , iee++)
       {
         if (j > ib)
         {
           j   += ib1 ;
           kkk += ib1 ;
         }

         ss  = *(a_p+iee)  ;
         kk  = kkk - 1     ;
         for (k = j-1 ; k < ie ; k += ib1 , kk += ib1)
           ss -= *(a_p+k) * *(a_p+kk) ;

         *(a_p+iee) = ss * s ;
       }

       ie          += ib ;
       *(a_p+ie-1) *= s ;

       fillin_f (50 , i , 0L) ;
     }
   }

   ie = ib1 * ib ;

   for (i = ib ; i <= n ; i++)
   {
     s  = *(a_p+ie) ;
     ia = ie - irue ;

     for (j = ia-1 ; j < ie ; j += ib1)
       s -= *(a_p+j) * *(a_p+j) ;

     if (s <= 0.) goto NonPosDef ;

     if (i == 2)
       smin = s ;
     else
     {
       if (s < smin * eps) goto NonPosDef ;
       if (s < smin)       smin = s ;
     }

     s = sqrt (s)  ;
     *(a_p+ie) = s ;
     s = 1. / s   ;

     if (ib != 2)
     {
       iaa = ia + ib ;
       iee = ie + 1  ;

       for (j = iaa ; j <= ie ; j += ib , iee++)
       {
         ia  += ib1       ;
         ss  = *(a_p+iee) ;
         kk  = j - 1      ;

         for (k = ia-1 ; k < ie ; k += ib1 , kk += ib1)
           ss -= *(a_p+k) * *(a_p+kk) ;

         *(a_p+iee) = ss * s ;
       }
     }

     ie          += ib ;
     *(a_p+ie-1) *=  s ;
                                   
     fillin_f (50 , i , 0L) ;
   }
      
   PosDef:
            succ = 1 ;
               
   NonPosDef:         
      
   fillin_f (1000 , 0L , 0L) ;
      
   return succ ;
}
