/* ************************************************************************* */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                               */
/*                                                                           */
/*  Konvertierung des li_p-Feldes (identische Verschiebungen) vom            */
/*  Eingabeformat in das Verarbeitungsformat                                 */
/*  ==============================================================           */
/*                                                                           */
/*  *li_p ist ein Vektor mit nk*kf Elementen, der auf den ersten             */
/*  nk Positionen die Informationen fuer alle 1. Freiheitsgrade,             */
/*  auf den zweiten nk Positionen die Informationen fuer alle                */
/*  2. Freiheitsgrade usw. enthaelt, er wird auf knotenweise                 */
/*  Information umgebaut durch "Transponieren".                              */
/*                                                                           */
/*  Im EINGABEFORMAT sind alle Freiheitsgrade, deren Verschiebungen          */
/*  identisch sein sollen, durch die (beliebige) gleiche ganze Zahl          */
/*  gekennzeichnet (es gibt "Gruppen", die jeweils durch die gleiche         */
/*  Zahl gekennzeichnet sind).                                               */
/*                                                                           */
/*  Im VERARBEITUNGSFORMAT wird "gepointert": Alle Freiheitsgrade            */
/*  einer Gruppe "pointern" auf einen Freiheitsgrad, dem "Gruppen-           */
/*  mitglied", das den groessten Index im Verschiebungsvektor hat,           */
/*  das selbst eine 0 im li_p-Vektor bekommt und sich bei der                */
/*  Verarbeitung deshalb "passiv" verhaelt.                                  */
/*                                                                           */
/*  Diese "Pointer" im Verarbeitungsformat entsprechen den (mit 0            */
/*  beginnenden) Indizes des Verschiebungsvektors. Weil eine 0               */
/*  das Kennzeichen fuer "Nicht-Pointern" ist, koennte nicht auf             */
/*  die Position 0 gepointert werden, was aber ohnehin nicht                 */
/*  vorkommen kann, da grundsaetzlich "vorwaerts gepointert" wird.           */
/*                                                                           */
/*  Die Funktion restli_f, die nach dem Abarbeiten des FEM_Algorithmus       */
/*  aufgerufen wird, macht die von convli_f veraenderte Anordnung wieder     */
/*  rueckgaengig.                                                            */
/*                                                                           */
/*  Eingabe:   li_p  -  Identische Verschiebungen                            */
/*             nk    -  Knotenanzahl                                         */
/*             kf    -  Maximale Anzahl der Freiheitsgrade pro Knoten        */
/*                                                                           */
/*  Autor: J. Dankert                                                        */
/* ************************************************************************* */

#include    "femset_f.h"
#include  "w_fs_res.h"

int convli_f (long* li_p , long nk , long kf)
{
  long i , j , j1 , nfg ;
  long *lisave_p ;                                          /*  Hilfsfeld    */

  nfg = nk * kf ;

  for (i = 0 ; i < nfg ; i++) if (*(li_p + i) != 0) break ;

  if (i >= nfg) return 1 ;                      /*  Es gibt nichts zu tun    */

  if ((lisave_p = (long*) calloc_f (nfg , sizeof (long))) == NULL)
  {
    infbox_f (S_SORRY_f , S_CONVLISPACE_f) ;
    return 0 ;
  }

  /*  "Transponieren" der li_p-Matrix:                                       */

  for (i = 0 ; i < nk ; i++)
    for (j = 0 ; j < kf ; j++) 
    {
      j1                     = j * nk + i   ;
      *(lisave_p + i*kf + j) = *(li_p + j1) ;
      *(li_p     + j1      ) = 0            ;
    }

  /*  ... und auf lisave_p steht die Eingabe-Information in anderer          */
  /*      Reihenfolge (knotenweise).                                         */

  /*  Erzeugen des Verarbeitungsformats:                                     */

  for (i = nfg - 1 ; i > 0 ; i--)      /*  Durchsuchen beginnt am Ende       */
    if (*(lisave_p + i) != 0) 
    {
      j1 = *(lisave_p + i) ;           /*  Kandidat mit "Gruppennummer" j1   */
      for (j = i - 1 ; j >= 0 ; j--)
        if (*(lisave_p + j) == j1)
        {                              /*  Gruppenmitglieder bekommen        */
          *(li_p     + j) = i ;        /*  Position i als "Pointer"          */
          *(lisave_p + j) = 0 ;
        }
    }

    freesp_f (lisave_p) ;              /*  Speicherplatz freigeben           */
    return 1 ;
}
