/* ************************************************************************* */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                               */
/*                                                                           */
/*  Aufbau der Elementsteifigkeitsbeziehung                                  */
/*  =======================================                                  */
/*                                                                           */
/*  Die uebergebene Elementnummer ie ist die Nummer der "Benutzer-           */
/*  Numerierung" (1 ... ne) der Elemente, in km_p stehen ebenfalls die       */
/*  Knotennummern der "Benutzer-Numerierung" (1 ... nk).                     */
/*                                                                           */
/*  Die Felder sind allerdings (wie in C ueblich) mit der unteren Index-     */
/*  grenze 0 definiert, auf  km [3] [0]  steht also z. B. die Knotennummer   */
/*  des ersten zum Element 4 gehoerenden Elementknotens.                     */
/*                                                                           */
/*  km_p enthaelt in jeder Zeile die Knotennummern fuer ein Element, alle    */
/*  Zeilen haben die Laenge kemx (auch dann, wenn Elemente weniger als       */
/*  kemx Knoten haben).                                                      */
/*                                                                           */
/*  elemat_f wertet die Matrix et_p (Elementtyp) aus, die in einer Zeile ie  */
/*  5 Informationen fuer das Element (ie+1) enthaelt:                        */
/*                                                                           */
/*      *  Elementtyp,                                                       */
/*      *  Anzahl der Knoten,                                                */
/*      *  Anzahl der Elementparameter,                                      */
/*      *  Dimension des Elements (ein-, zwei- oder dreidimensional),        */
/*      *  Anzahl der Freiheitsgrade pro Knoten                              */
/*                                                                           */
/*  (entspricht einer kompletten "individuellen Element-Charakteristik"      */
/*  fuer den Elementtyp).                                                    */
/*                                                                           */
/*  Abhaengig vom Elementtyp wird die passende Funktion fuer den Aufbau      */
/*  der Elementsteifigkeitsbeziehung gestartet. Wenn der Elementtyp nicht    */
/*  bekannt ist (oder bei Elementtyp 0), wird die Funktion emstif_f          */
/*  gestartet (man kann also einen neuen Elementtyp zunaechst in der         */
/*  Funktion emstif_f definieren und testen).                                */
/*                                                                           */
/*  Wenn die Funktion emstif_f den Return-Wert 0 liefert (Default-Wert fuer  */
/*  die "Dummy-Funktion" in C-FEMSET), wird noch ein letzter Versuch         */
/*  gestartet: Die elemat_f uebergebene Element-Charakteristik (ke = kemx,   */
/*  kf, kx, kp) wird mit den Element-Charakteristiken der bekannten          */
/*  Elementsteifigkeitsbeziehungen verglichen. Wenn sie als bekannt          */
/*  erkannt wird, wird die entsprechende Funktion gestartet. So wird auch    */
/*  Aufwaertskompatibilitaet zu aelteren FEMSET-Versionen, die nur jeweils   */
/*  einen Elementtyp zuliessen (und keine Matrix et_p kannten) hergestellt.  */
/*                                                                           */
/*  Die Funktion, die die Elementsteifigkeitsbeziehung aufbaut, kann mit     */
/*  relativen Pointer-Adressen oder mit Arrays arbeiten (Beipiel fuer        */
/*  das Arbeiten mit relativen Pointer-Adressen: rahm2d_f.c, Beispiel fuer   */
/*  das Arbeiten mit einem Array fuer die Elementsteifigkeitsmatrix:         */
/*  rahm3d_f.c), beim Definieren eines Arrays muss man gegebenenfalls        */
/*  eine Warnung wie "Suspicious Pointer Conversion" ignorieren.             */
/*                                                                           */
/*  Autor: J. Dankert                                                        */
/* ************************************************************************* */

#include    "femset_f.h"

int elemat_f (long    ie   ,
              double *xy_p , long   *km_p , long   *et_p , double *ep_p ,
              long    kemx , long    kx   , long   *kf   , long   *ke   ,
              long    kp   , double *em_p , double *eb_p)
{
  long  eltyp , i ;

  eltyp = *(et_p + (ie-1) * 5) ;

  for (i = 1 ; i <= 2 ; i++)                    /* Gegebenenfalls 2 Versuche */
  {
    if (eltyp == FSET_TRUSS2D_f)
      return (fach2d_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_TRUSS2DT_f)
      return (fact2d_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_BEND2D_f)
      return (bieg2d_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_BD2DEB_f)
      return (bg2deb_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_BD2DKM_f)
      return (bg2dkm_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_BEAM2D_f)
      return (rahm2d_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_BM2DKM_f)
      return (rm2dkm_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_TRUSS3D_f)
      return (fach3d_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_BEAM3D_f)
      return (rahm3d_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_PLANE3_6_f)
      return (sd6_f    (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_PLANE4_8_f)
      return (sv8_f    (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_PLANE4_16_f)
      return (sv16_f   (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else if (eltyp == FSET_TORS3_6_f)
      return (torsd6_f (ie , xy_p , km_p , ep_p , kemx ,
                        kx , kf   , ke   , kp   , em_p , eb_p)) ;
    else
    {
      /* "Dummy" emstif_f koennte fuer einen neuen Elementtyp benutzt 
         werden:                                                             */
      if (emstif_f (ie , xy_p , km_p , ep_p , kemx ,
                    kx , kf   , ke   , kp   , em_p , eb_p)) return 1 ;    
                      
      if (eltyp == 0)
        { 
          /* Versuch, den Elementtyp aus der Element-Charakteristik
             zu ermitteln:                                                   */
          if (!(eltyp = elmtyp_f (kx , *kf , *ke , kp))) return 0 ;
        }  
    }
  }
  return 0 ;
}
