/* ************************************************************************* */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                               */
/*                                                                           */
/*  Finite-Elemente-Algorithmus                                              */
/*  ===========================                                              */
/*                                                                           */
/*  Es wird der komplette Finite-Elemente-Algotrithmus realisiert mit        */
/*                                                                           */
/*     *  Aufbau der Systemsteifigkeitsbeziehung,                            */
/*     *  Einbau der geometrischen Randbedingungen und                       */
/*     *  Loesung des linearen Gleichungssystems (Cholesky-Verfahren).       */
/*                                                                           */
/*  Eingabe:   kx    -  "Dimension" des Problems (1, 2 oder 3)               */
/*             kf    -  Maximale Anzahl der Freiheitsgrade pro Knoten        */
/*             ke    -  Maximale Knotenanzahl pro Element                    */
/*             kp    -  Anzahl der Parameter, die ein Element beschreiben    */
/*             ne    -  Elementanzahl                                        */
/*             nk    -  Knotenanzahl                                         */
/*             xy_p  -  Knotenkoordinaten                                    */
/*             km_p  -  Koinzidenzmatrix                                     */
/*             et_p  -  Element-Typen                                        */
/*             ep_p  -  Elementparameter                                     */
/*             kr_p  -  Randbedingungen                                      */
/*             bk_p  -  Knotenlasten                                         */
/*             sc_p  -  Federsteifigkeiten                                   */
/*             la_p  -  Lager unter Winkel                                   */
/*             aa_p  -  Winkel fuer Lager                                    */
/*             li_p  -  Identische Verschbungen                              */
/*             pu_p  -  Vorgeschrbiebene Verschiebungen                      */
/*                                                                           */
/*  Return-Wert:  1  --->  Erfolg, die berechneten Verschiebungen wurden     */
/*                         auf dem Feld uv_p abgelegt                        */
/*                0  --->  Misserfolg                                        */
/*                                                                           */
/*  Autor: J. Dankert                                                        */
/* ************************************************************************* */

#include    <time.h>
#include    "femset_f.h"
#include    "w_fs_res.h"

int femalg_f (long    kx   , long    kf   , long    ke   ,
              long    kp   , long    ne   , long    nk   ,
              double *xy_p , long   *km_p , long   *et_p , double *ep_p ,
              long   *kr_p , double *bk_p , double *sc_p , long   *la_p ,
              double *aa_p , long   *li_p , double *pu_p , double *uv_p)
{
  double   *a_p , *b_p ;                      /* Systemsteifigkeitsbeziehung */
  long      suc ,  i   ;

  if (!convli_f (li_p , nk , kf)) return 0 ;  /*  ... konvertiert li_p aus
                                        Eingabeformat in Verarbeitungsformat */

  ib_f = dimsys_f (km_p , et_p , li_p , ne , ke , kf , nk , &n_f) ;
  
    /*  ... ermittelt die Bandweite ib_f und die Anzahl der Gleichungen 
            n_f fuer die Systemsteifigkeitsbeziehung                         */

  /****  Speicherplatz fuer Systemsteifigkeitsbeziehung allokieren: **********/

  protoc_f (FSPR_STARTALLOC_f) ;                      /* Protokollausschrift */

  suc = ((a_p = (double *) calloc_f (n_f * ib_f ,
                                     sizeof (double))) != NULL) ;
  if (suc) suc = ((b_p = (double *) calloc_f (n_f ,
                                     sizeof (double))) != NULL) ;
  if (!suc)
  {
    infbox_f (S_SORRY_f , S_NOSTIFFSPACE_f) ;
    return 0 ;
  }

  /****  Aufbau und Loesung des Gleichungssystems:  **************************/

  darcop_f (bk_p , b_p , n_f) ;        /*  Knotenlasten --> Belastungsvektor */
    
  lestrc_f (n_f , &ib_f , FMT_LEQSTRC_f) ;              /* Struktur zeichnen */
  protoc_f (FSPR_STARTSYSTEM_f) ;
    
  StartTime_f = (double) clock () / CLOCKS_PER_SEC ;

  if (!system_f (a_p  , b_p  , n_f  , ib_f , ne  , nk ,   /* Aufbau der      */
                  kx   , kf   , ke   , kp   ,             /* Systemsteifig-  */
                  xy_p , km_p , et_p , ep_p , li_p))      /* keitsmatrix     */
  {
    infbox_f (S_SORRY_f , S_STIFFERROR_f) ;
    return 0 ;
  }   

  StiffTime_f = (double) clock () / CLOCKS_PER_SEC - StartTime_f ;
  protoc_f (FSPR_ENDSYSTEM_f) ;

  if (kx == 2) boubet_f (a_p , b_p , ib_f , nk , kf , la_p , aa_p) ;
      
            /* ... baut verhinderte Verschiebungen ein, die nicht mit einer  */
            /*     Koordinatenrichtung uebereinstimmen (nur 2D-Probleme!)    */

  boucon_f (a_p , b_p , ib_f , kr_p , nk , kf , li_p , pu_p) ;
  
            /* ... baut behinderte und vorgeschriebene Verschiebungen ein    */

  spccon_f (a_p , ib_f , sc_p , nk , kf) ;
  
            /* ... baut die speziellen Randbedingungen (Federsteiigkeiten)
                   in die System-Steifigkeitsbeziehung ein                   */

  protoc_f (FSPR_STARTCHOL_f) ;

  StartTime_f = (double) clock () / CLOCKS_PER_SEC ;

  if (!cholba_f (a_p , n_f , ib_f , lineq_eps))      /*  Cholesky-Zerlegung  */
  {    
    infbox_f (S_SORRY_f , S_NOTPOSDEF_f) ;
    freesp_f (a_p) ;
    freesp_f (b_p) ;
    return 0 ;
  }

  SolveTime_f = (double) clock () / CLOCKS_PER_SEC - StartTime_f ;
  protoc_f (FSPR_ENDCHOL_f) ;

  /*  Vorwaerts-Rueckwaerts-Einsetzen (Knotenverschiebungen berechnen):      */
  
  vorrue_f (a_p , b_p , uv_p , n_f , ib_f , 1L) ;

  /*  Gleichsetzen der "verweisenden" Freiheitsgrade:                        */
  
  for (i = 0 ; i < nk * kf ; i++)
  {
    if (*(li_p+i) > i) *(uv_p+i) = *(uv_p + *(li_p+i)) ; 
  }

  if (kx == 2) bouret_f (uv_p , nk , kf , la_p , aa_p) ;
  
          /*  ... rechnet verhinderte Verschiebungen zurueck, die nicht
                  mit einer Koordinatenrichtung uebereinstimmen              */

  restli_f (li_p , nk , kf) ;        /*  ... konvertiert li_p wieder in das  */
                                     /*      Eingabeformat zurueck           */
  freesp_f (a_p) ;
  freesp_f (b_p) ;
  return 1 ;
}
