/* ************************************************************************* */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                               */
/*                                                                           */
/*  Finite-Elemente-Algorithmus                                              */
/*  ===========================                                              */
/*                                                                           */
/*  Es wird der komplette Finite-Elemente-Algotrithmus fuer die Berechnung   */
/*  von Eigenschwingungen realisiert mit                                     */
/*                                                                           */
/*     *  Aufbau der Systemsteifigkeitsmatrix,                               */
/*     *  Aufbau der Systemmassenmatrix,                                     */
/*     *  Einbau der geometrischen Randbedingungen und                       */
/*     *  Loesung des Matizeneigenwertproblems (inverse Simultaniteration).  */
/*                                                                           */
/*  Eingabe:   neig  -  Anzahl der zu berechnenden Eigenfrequenzen           */
/*             kx    -  "Dimension" des Problems (1, 2 oder 3)               */
/*             kf    -  Maximale Anzahl der Freiheitsgrade pro Knoten        */
/*             ke    -  Maximale Knotenanzahl pro Element                    */
/*             kp    -  Anzahl der Parameter, die ein Element beschreiben    */
/*             ne    -  Elementanzahl                                        */
/*             nk    -  Knotenanzahl                                         */
/*             xy_p  -  Knotenkoordinaten                                    */
/*             km_p  -  Koinzidenzmatrix                                     */
/*             et_p  -  Element-Typen                                        */
/*             ep_p  -  Elementparameter                                     */
/*             kr_p  -  Randbedingungen                                      */
/*             bk_p  -  Knotenmassen                                         */
/*             sc_p  -  Federsteifigkeiten                                   */
/*             la_p  -  Lager unter Winkel                                   */
/*             aa_p  -  Winkel fuer Lager                                    */
/*             li_p  -  Identische Verschbungen                              */
/*             pu_p  -  Vorgeschrbiebene Verschiebungen                      */
/*                                                                           */
/*  Return-Wert: nc  --->  Erfolg, nc berechnete Eigenwerte wurden auf       */
/*                         dem Feld ev_p, die zugehoerigen Eigenvektoren     */
/*                         auf dem Feld x_p abgelegt (nc <= neig)            */
/*                0  --->  Misserfolg                                        */
/*                                                                           */
/*  Autor: J. Dankert                                                        */
/* ************************************************************************* */

#include    <time.h>
#include    "femset_f.h"
#include    "w_fs_res.h"

int femeig_f (long    neig , long    kx   , long    kf   , long    ke   ,
              long    kp   , long    ne   , long    nk   ,
              double *xy_p , long   *km_p , long   *et_p , double *ep_p ,
              long   *kr_p , double *bk_p , double *sc_p , 
			  double *ev_p , double *x_p)
{
  double   *a_p , *b_p ;                      /* Systemsteifigkeitsbeziehung */
  long      suc ,  i   , ii , nc , itmax , nred ;

  ib_f = dimeig_f (km_p , et_p , ne , ke , kf , nk , &n_f) ;
  
    /*  ... ermittelt die Bandweite ib_f und die Anzahl der Gleichungen 
            n_f fuer die Matrizen a_p und b_p                                */

  /****  Speicherplatz fuer Matrizen a_p und b_p allokieren: *****************/

  protoc_f (FSPR_STARTALLOCEWP_f) ;                   /* Protokollausschrift */

  suc = ((a_p = (double *) calloc_f (n_f * ib_f ,
                                     sizeof (double))) != NULL) ;
  if (suc) suc = ((b_p = (double *) calloc_f (n_f *ib_f ,
                                     sizeof (double))) != NULL) ;
  if (!suc)
  {
    infbox_f (S_SORRY_f , S_NOSTIFFSPACE_f) ;
    return 0 ;
  }

  /****  Aufbau und Loesung des Matrizeneigenwertproblems:  ******************/

  for (i = 0 , ii = 0 ; i < n_f ; i++ , ii += ib_f)
	  *(b_p + ii) = *(bk_p + i) ;              /*  Knotenmassen --> B-Matrix */
    
  lestrc_f (n_f , &ib_f , FMT_LEQSTRC_f) ;              /* Struktur zeichnen */
  protoc_f (FSPR_STARTEWPMATRIZEN_f) ;
    
  StartTime_f = (double) clock () / CLOCKS_PER_SEC ;

  if (!syseig_f (a_p  , b_p  , n_f  , ib_f , ne  , nk ,   /* Aufbau der      */
                 kx   , kf   , ke   , kp   ,              /* Matrizen A      */
                 xy_p , km_p , et_p , ep_p))              /* und B           */
  {
    infbox_f (S_SORRY_f , S_STIFFERROR_f) ;
    return 0 ;
  }   

  StiffTime_f = (double) clock () / CLOCKS_PER_SEC - StartTime_f ;
  protoc_f (FSPR_ENDEWPMATRIZEN_f) ;

  spccon_f (a_p , ib_f , sc_p , nk , kf) ;
  
            /* ... baut die speziellen Randbedingungen (Federsteiigkeiten)
                   in die System-Steifigkeitsbeziehung ein                   */

  nred = lcdeig_f (a_p , b_p , ib_f , kr_p , nk , kf) ;
  
            /* ... baut behinderte Verschiebungen ein                        */

  protoc_f (FSPR_STARTISIASB_f) ;

  StartTime_f = (double) clock () / CLOCKS_PER_SEC ;

  itmax = isiasb_imx ;
  if (isiasb_f (a_p , b_p , nred , ib_f , ib_f , neig , 
	            &itmax , isiasb_eps , &nc  , ev_p , x_p) != 1)
  {    
    infbox_f (S_SORRY_f , S_NOTPOSDEF_f) ;
    freesp_f (a_p) ;
    freesp_f (b_p) ;
    return 0 ;
  }

  Iteration_f = itmax ;
  SolveTime_f = (double) clock () / CLOCKS_PER_SEC - StartTime_f ;
  protoc_f (FSPR_ENDISIASB_f) ;

  expvec_f (x_p , ib_f , kr_p , nk , kf , nred , nc) ;

  freesp_f (a_p) ;
  freesp_f (b_p) ;
  return nc ;
}
