/* ************************************************************************* */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                               */
/*                                                                           */
/*  Gram-Schmidt-Orthogonalisierung der Spalten der Matrix X                 */
/*  ========================================================                 */
/*                                                                           */
/*  Die Matrix X (m Zeilen und neiwe Spalten) wird nach Gram-Schmidt         */
/*  orthonormiert.                                                           */
/*                                                                           */
/*  B ist eine symmetrische Bandmatrix mit m Zeilen und der Bandweite ib,    */
/*  von der die wesentlichen m*ib Elemente in einem eindimensionalen Feld    */
/*  zeilenweise dicht gespeichert sind.                                      */
/*                                                                           */
/*  Auf dem Feld R (pointer r_p) werden die "Reinigungskoeffizienten" rik    */
/*  abgeliefert. Es ist eine spaltenweise dicht gepackte Dreiecksmatrix      */
/*  mit (neiwe+1)*neiwe/2 Elementen. R konvergiert gegen die                 */
/*  Diagonalmatrix der Eigenwerte.                                           */
/*                                                                           */
/*  Autor: J. Dankert                                                        */
/* ************************************************************************* */

#include <math.h>
#include    "femset_f.h"

void grasch_f (double *x_p , double *b_p , double *r_p  , 
				 long    m   , long   ib , long   *neiwe)
{
  long   i   , j , k , ihdr , ijr , ii , iia , jj , kend ,
	     iii , jjj , iijj , cont ;
  double rkk , s ;

  cont = 1 ;

  while (cont)
  {
    cont = 0 ;
    ihdr = 0 ;
    ijr  = 0 ;
  
    /* Schleife ueber alle Spalten von X: */
    for (k = 0 ; k < *neiwe ; k++)
    {
    	  /* Erste Spalte wird nur normiert: */
    	  if (k > 0)
		  {
             /* Schleife ueber die Spalten  0 ... k-1  zur Ermittlung
                der "Reinigungskoeffizienten" rik (j ist die Position
                des Elements rik): */
             for (j = ijr , jj = 0 ; j <= ihdr ; j++ , jj++)
               *(r_p+j) = ytbibx_f (x_p , b_p , m , ib , *neiwe , jj , k) ;

             /* "Reinigung" der k-ten Spalte von X von
                den Komponenten der Spalten  0 ... k-1: */
             kend = k + (m-1) * *neiwe ;
             for (j = ijr , iia = 0 ;  j <= ihdr ; j++ , iia++)
			 {
			     s = *(r_p+j) ;
                 for (i = k , ii = iia ; i <= kend ; i += *neiwe , ii += *neiwe)
                     *(x_p+i) -= *(x_p+ii) * s ;
			 }

             ihdr++ ;
		  }
	  
	      /* Normierung: */
          rkk = ytbibx_f (x_p , b_p , m , ib , *neiwe , k , k) ;
	      /* ... liefert den Hauptdiagonal-Normierungskoeffizienten rkk */
	      rkk = sqrt (rkk) ;
	      *(r_p + ihdr) = rkk ;

	      /* Normierung der k-ten Spalte: */
	      if (fabs(rkk) > grasch_eps)
		  {
		      rkk = 1./rkk ;
              for (j = 0 , jj = k ; j < m ; j++ , jj += *neiwe)
			      *(x_p + jj) *= rkk ;

		      ihdr += k+1 ;   /* Position in R vor dem naechsten Hauptdiagonalelement */
		      ijr  += k+1 ;   /* Anfangsposition naechste Spalte in R */
		  }
	      else
		  {
		      /* ... besteht Ueberlaufgefahr, was eigentlich nur
                     bei singulaerer B-Matrix passieren kann. Es wird
                     versucht, durch Reduktion der Anzahl der zu
                     berechnenden Eigenwerte (und -vektoren) wenigstens
                     die berechenbaren Eigenwerte zu ermitteln.
                     Hier verwendeter "Trick": Kritischer Eigenvektor
                     wird ausgesondert und NEIWE wird reduziert. */
              if (k+1 < *neiwe)
			  {
                for (jjj = k+1 ; jjj < *neiwe ; jjj++)
                     for (iii = 0 , iijj = jjj ; iii < m ; iii++ , iijj += *neiwe)
					     *(x_p + iijj - 1) = *(x_p + iijj) ;
			  }

		      (*neiwe)-- ;
		      if (*neiwe > 0)
			  {
			      cont = 1 ;
			      break ;
			  }
		      else
			      return ;
		  }
	} /* Ende  for k=0 ...  */
  }   /* Ende  while (cont) */

  return ;
}
