/* ************************************************************************* */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                               */
/*                                                                           */
/*  Eigenwerte und Eigenvektoren des allgemeinen Eigenwertproblems mit       */
/*  symmetrischen Bandmatrizen (Gram-Schmidt-Iteration)                      */
/*  ==================================================================       */
/*                                                                           */
/*  Es werden die neiwe kleinsten Eigenwerte und die zugehoerigen            */
/*  Eigenvektoren berechnet (und auf den Adressen eiwe_p bzw. x_p            */
/*  abgeliefert). Die Matrix A des allgemeinen symmetrischen                 */
/*  Eigenwertproblems (A - lambda * B) x = o muss positiv definit sein.      */
/*                                                                           */
/*  Eingabe:  a_p - Symmetrische Bandmatrix A mit m Zeilen und der Bandweite */
/*                  iba, gespeichert sind zeilenweise die m*iba              */
/*                  wesentlichen Elemente (zeilenweise in einem              */
/*                  eindimensionalen Feld)                                   */
/*            b_p - Symmetrische Bandmatrix B mit m Zeilen und der Bandweite */
/*                  ibb, gespeichert sind zeilenweise die m*ibb              */
/*                  wesentlichen Elemente (zeilenweise in einem              */
/*                  eindimensionalen Feld)                                   */
/*            m   - Zeilenanzahl der Matrizen A und B                        */
/*            iba - Bandweite der Matrix A                                   */
/*            ibb - Bandweite der Matrix B                                   */
/*          neiwe - Anzahl der zu berechnenden Eigenwerte                    */
/*          itmax - Maximale Anzahl der auszufhrenden Iterationen           */
/*            eps - Relativer tolerierter Fehler fuer die Eigenwerte         */
/*                                                                           */
/* Ergebnisse:                                                               */
/*                                                                           */
/*        nkonv_p - Adresse, auf der die Anzahl der tatsaechlich             */
/*                  berechneten Eigenwerte abgeliefert wird (ist im          */
/*                  Erfolgsfall gleich neiwe, nur bei vorzeitigem            */
/*                  Abbruch wegen Erreichens von itmax Iterationen kann      */
/*                  *nkonv_p kleiner als neiwe sein)                         */
/*        itmax_p - Adresse, auf der die tatsaechlich ausgefuehrte Anzahl    */
/*                  von Iterationen zurueckgegeben wird (Wert ist negativ,   */
/*                  wenn die Anzahl der maximal auszufuehrenden Iterationen  */
/*                  erreicht wurde)                                          */
/*         eiwe_p - Vektor mit neiwe Elementen, enthaelt die berechneten     */
/*                  Eigenwerte                                               */
/*            x_p - Matrix mit m Zeilen und neiwe Spalten, enthaelt die      */
/*                  berechneten Eigenvektoren                                */
/*                                                                           */
/* Return-Wert:     1, wenn Berechnung ausgefuehrt wurde, 0, wenn Matrix     */
/*                  A nicht positiv definit ist, -1 bei einem                */
/*                  anderen Fehler (z. B.: Kein ausreichender                */
/*                  Arbeitsspeicherplatz)                                    */
/*                                                                           */
/*  Autor: J. Dankert                                                        */
/* ************************************************************************* */

#include    <math.h>
#include    "femset_f.h"

int isiasb_f (double *a_p     , double *b_p     , 
			  long   m        , long   iba      , long   ibb ,
			  long   neiw     , long   *itmax_p , double eps , 
			  long   *nkonv_p , double *eiwe_p  , double *x_p)
{
  long    i , j , ii , kon , neiwe , it = 0 , iter = 1 ,
	      iiba , iibb ;
  double *r_p , *aa_p , *bb_p , *as_p , ew ;

  /* Speicherplatz fuer Sicherung der Matrix A allokieren: */
    if ((aa_p = (double *) calloc_f (m*iba ,
                                     sizeof (double))) == NULL)
    {
    	  *nkonv_p = 0 ;
    	  return -1 ;
    }
	else
	{
      /* Sicherung der Matrix A: */
      for (i = 0 ; i < m*iba ; i++)
  	      *(aa_p + i) = *(a_p + i) ;
	}

  as_p  = aa_p ;
  bb_p  = b_p  ;
  iiba  = iba  ;
  iibb  = ibb  ;
  neiwe = neiw ;

  for (i = 0 ; i < neiwe ; i++)
  	  *(eiwe_p + i) = 0. ;

    /* Cholesky-Zerlegung der Matrix A: */
    if (!cholba_f (aa_p , m , iba , lineq_eps))
    {
		/* Matrix A nicht positiv definit: */
    	*nkonv_p = 0 ;
        freesp_f (aa_p) ;
    	return 0 ;
    }
    
    /* Workspace allokieren: */
    if ((r_p = (double *) calloc_f (m*neiwe ,
                                    sizeof (double))) == NULL)
    {
    	  *nkonv_p = 0 ;
          freesp_f (as_p) ;
    	  return -1 ;
    } 

    /* Startvektoren: */
    ii = 0 ;
    for (i = 0 ; i < m ; i++)
      for (j = 0 ; j < neiwe ; j++)
	  {
	       	*(x_p + ii++) = ii + 1 ;
	  }

  while (iter)
  {
    /************  Iterationszyklus, Beginn:  **********************/
    /* Multiplikation B*X erzeugt rechte Seite fuer 
       Iterationsvorschrift:                                       */
    matmub_f (bb_p , x_p , r_p , m , iibb  , neiwe) ;

    /* Inverse Iteration und Orthonormierung:                      */
    vorrue_f (aa_p , r_p  , x_p , m , iiba ,  neiwe) ;
    grasch_f (x_p  , bb_p , r_p , m , iibb , &neiwe) ;
    if (neiwe <= 0)
	{
      freesp_f (r_p)  ;
      freesp_f (as_p) ;
	  *nkonv_p = 0 ;
      return -1 ;
	}

    it++ ;

    /* Uebertragen der Eigenwerte auf das Feld eiwe_p,
       Pruefen auf Konvergenz:                                     */
    *nkonv_p = 0 ;
	kon      = 1 ;
    for (i = 0 , j = 0 ; i < neiwe ; i++)
	{
       ew = 1./(*(r_p+j)) ;
	   if (fabs((*(eiwe_p + i)-ew)/ew) > eps) kon = 0 ;
	   if (kon > 0) (*nkonv_p)++ ;
	   *(eiwe_p + i) = ew ;
	   j += i + 2 ;
	}

    if (*nkonv_p < neiwe)
	{
	    if (it > *itmax_p)
		{
	    	  *itmax_p = -it ;
			  iter     = 0   ;
		}
	}
    else
	{
	    *itmax_p = it ;
		iter     = 0  ;
	}
  }

  evnorm_f (x_p , m , *nkonv_p) ;

  freesp_f (r_p)  ;
  freesp_f (as_p) ;

  return 1 ;
}
