/* ************************************************************************* */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                               */
/*                                                                           */
/*  Aufbau der Elementsteifigkeitsbeziehung 'RAHMEN3D'                       */
/*  ==================================================                       */
/*                                                                           */
/*  Die uebergebene Elementnummer ie ist die Nummer der "Benutzer-           */
/*  Numerierung" (1 ... ne) der Elemente, in km_p stehen ebenfalls die       */
/*  Knotennummern der "Benutzer-Numerierung" (1 ... nk).                     */
/*                                                                           */
/*  Die Felder sind allerdings (wie in C ueblich) mit der unteren Index-     */
/*  grenze 0 definiert, auf km_p [3] [0] steht also z. B. die Knotennummer   */
/*  des ersten zum Element 4 gehoerenden Elementknotens (km_p ist mit ke     */
/*  Spalten vereinbart, was einen Zugriff auf dasselbe Element ueber         */
/*  *(km_p + kemx*3) zulaesst.                                               */
/*                                                                           */
/*  Autor: J. Dankert                                                        */
/* ************************************************************************* */

#include    <math.h>
#include    "femset_f.h"

#define     EPS  0.1e-6

int rahm3d_f (long ie   , double *xy_p , long   *km_p , double *ep_p ,
              long kemx , long    kx   , long   *kf   , long   *ke   ,
              long kp   , double *em_p , double *eb_p)
{
  long     k1 , k2 , i  , j  , k ;
  double   e , g , iyy , izz , it , a , ea , eiy , eiz , git ,
           x1 , y1 , z1 , x2 , y2 , z2 , l , ls ,
           cxx , cxy , cxz , cyx , cyy , cyz ,
           czx , czy , czz ,
           t[12][12] , tt[12][12] , h[12][12] , em[12][12] ;

  *ke = 2 ;                           /*  Knoten pro Element                 */
  *kf = 6 ;                           /*  Freiheitsgrade pro Knoten          */

  for (i = 0 ; i < 12 ; i++)
    for (j = 0 ; j < 12 ; j++) 
    {
      em [i] [j] = 0. ;
      t  [i] [j] = 0. ;
    }

  k1 = *(km_p + (ie-1)*kemx)     ;    /*  Knotennummern des Elements         */
  k2 = *(km_p + (ie-1)*kemx + 1) ;

  e   = *(ep_p + (ie-1)*kp)     ;     /*  Elastizitaetsmodul                 */
  g   = *(ep_p + (ie-1)*kp + 1) ;     /*  Gleitmodul                         */
  iyy = *(ep_p + (ie-1)*kp + 2) ;     /*  Flaechentraegheitsmoment Iyy       */
  izz = *(ep_p + (ie-1)*kp + 3) ;     /*  Flaechentraegheitsmoment Izz       */
  it  = *(ep_p + (ie-1)*kp + 4) ;     /*  Torsionstraegheitsmoment           */
  a   = *(ep_p + (ie-1)*kp + 5) ;     /*  Querschnittsflaeche                */

  ea  = e * a   ;                     /*  Dehnsteifigkeit                    */
  eiy = e * iyy ;                     /*  Biegesteifigkeit bezuegl. y-Achse  */
  eiz = e * izz ;                     /*  Biegesteifigkeit bezuegl. z-Achse  */
  git = g * it  ;                     /*  Torsionssteifigkeit                */

  x1 = *(xy_p + (k1-1)*kx)     ;      /*  Knotenkoordinaten                  */
  y1 = *(xy_p + (k1-1)*kx + 1) ;
  z1 = *(xy_p + (k1-1)*kx + 2) ;
  x2 = *(xy_p + (k2-1)*kx)     ;      /*  Knotenkoordinaten                  */
  y2 = *(xy_p + (k2-1)*kx + 1) ;
  z2 = *(xy_p + (k2-1)*kx + 2) ;

  l  = sqrt ((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1)
                                   + (z2 - z1) * (z2 - z1)) ;
  if (l < 1.e-8) 
  {
    sprintf (WorkString_f , "Element%6ld hat die Lnge%12.6g" , ie , l) ;
    infbx2_f ("Element wird nicht verwendet!" , WorkString_f) ;
    return 0 ;
  }

  cxx = (x2 - x1) / l ;               /*  Richtungskosinusse                 */
  cxy = (y2 - y1) / l ;
  cxz = (z2 - z1) / l ;

  ls  = sqrt (cxx * cxx + cxy *cxy) ;

  if (ls < EPS) 
  {
    cyx = 0. ;                 /* Traegerachse parallel zur globalen         */
    cyy = 1. ;                 /* z-Achse, lokale "y-quer"-Achse liegt       */
    cyz = 0. ;                 /* parallel zur globalen y-Achse              */
  }
  else 
  {
    cyx = - cxy / ls ;         /* Allgemeine Lage der Traegerachse, lokale   */
    cyy =   cxx / ls ;         /* "y-quer"-Achse liegt parallel zur          */
    cyz =   0.       ;         /* globalen x-y-Ebene                         */
  }

  czx = cxy * cyz - cyy * cxz ;         /* cz-Werte als Vektorprodukt der    */
  czy = cyx * cxz - cxx * cyz ;         /* cx-Werte und der cy-Werte         */
  czz = cxx * cyy - cyx * cxy ;

  em [ 0] [ 0] =   ea  / l                 ;
  em [ 0] [ 6] = - em [0] [0]              ;
  em [ 1] [ 1] =   eiz * 12. / (l * l * l) ;
  em [ 1] [ 5] =   eiz *  6. / (l * l)     ;
  em [ 1] [ 7] = - em [1] [1]              ;
  em [ 1] [11] =   em [1] [5]              ;
  em [ 2] [ 2] =   eiy * 12. / (l * l * l) ;
  em [ 2] [ 4] = - eiy *  6. / (l * l)     ;
  em [ 2] [ 8] = - em [2] [2]              ;
  em [ 2] [10] =   em [2] [4]              ;
  em [ 3] [ 3] =   git / l                 ;
  em [ 3] [ 9] = - em [3] [3]              ;
  em [ 4] [ 4] =   eiy *  4. / l           ;
  em [ 4] [ 8] =   eiy *  6. / (l * l)     ;
  em [ 4] [10] =   eiy *  2. / l           ;
  em [ 5] [ 5] =   eiz *  4. / l           ;
  em [ 5] [ 7] = - em [1] [5]              ;
  em [ 5] [11] =   eiz *  2. / l           ;
  em [ 6] [ 6] =   em [0] [0]              ;
  em [ 7] [ 7] =   em [1] [1]              ;
  em [ 7] [11] = - em [1] [5]              ;
  em [ 8] [ 8] =   em [2] [2]              ;
  em [ 8] [10] = - em [2] [4]              ;
  em [ 9] [ 9] =   em [3] [3]              ;
  em [10] [10] =   em [4] [4]              ;
  em [11] [11] =   em [5] [5]              ;

  for (i = 0 ; i < 12 ; i++)
    for (j = i+1 ; j < 12 ; j++)
      em [j] [i] = em [i] [j] ;

  t [0] [0] = cxx ;
  t [0] [1] = cxy ;
  t [0] [2] = cxz ;
  t [1] [0] = cyx ;
  t [1] [1] = cyy ;
  t [1] [2] = cyz ;
  t [2] [0] = czx ;
  t [2] [1] = czy ;
  t [2] [2] = czz ;
  
  for (k = 3 ; k <= 9 ; k += 3)
    for (i = 0 ; i < 3 ; i++)
      for (j = 0 ; j < 3 ; j++)
        t [k+i] [k+j] = t [i] [j] ;

  matmul_f (&em [0] [0] , &t [0] [0] , &h [0] [0] , 12L , 12L , 12L) ;

  for (i = 0 ; i < 12 ; i++)
    for (j = 0 ; j < 12 ; j++)
      tt [i] [j] = t [j] [i] ;

  matmul_f (&tt [0] [0] , &h [0] [0] , em_p , 12L , 12L , 12L) ;

  return 1 ;
}
