/* *********************************************************** */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                 */
/*                                                             */
/*  Cholesky-Zerlegung einer Bandmatrix (Skyline-Verfahren)    */
/*  =======================================================    */
/*                                                             */
/*  Fuer die symmetrische, positiv definite Matrix A           */
/*  wird die Dreieckszerlegung nach Cholesky                   */
/*                                                             */
/*                           T                                 */
/*                      A = R  R                               */
/*                                                             */
/*  ausgefuehrt. Von A sind nur die wesentlichen Elemente      */
/*  in n eindimensionalen double-Arrays gespeichert:           */
/*                                                             */
/*  Jedes double-Array enthaelt die Werte eines "vertikalen    */
/*  Schornsteins", beginnend mit dem Hauptdiagonalelement,     */
/*  dem die darueber liegenden Elemente folgen. Die            */
/*  Anfangsadressen dieser Arrays (Adressen der Speicher-      */
/*  mit den Hauptdiagonalelementen) werden in einem Pointer-   */
/*  Array mdiag_p erwartet. Die Anzahl der darueber liegenden  */
/*  Elemente (ohne Hauptdiagonalelement) muessen in einem      */
/*  Array height angeliefert werden.                           */
/*                                                             */
/*  Beispiel:                                                  */
/*                                                             */
/*  Symmetrische Matrix:          Gespeichert werden...        */
/*                                                             */
/*    4  2  0  3  0  7   ...ab Adr. mdiag_p[0]:   4            */
/*    2  6  2  0  0  0   ...ab Adr. mdiag_p[1]:   6 2          */
/*    0  2  7  1  4  0   ...ab Adr. mdiag_p[2]:   7 2          */
/*    3  0  1  5  9  9   ...ab Adr. mdiag_p[3]:   5 1 0 3      */
/*    0  0  4  9  1  2   ...ab Adr. mdiag_p[4]:   1 9 4        */
/*    7  0  0  9  2  2   ...ab Adr. mdiag_p[5]:   2 2 9 0 0 7  */
/*                                                             */
/*  ...und das Array heigth muss folgende Werte enthalten:     */
/*                                                             */
/*                    0  1  1  3  2  5                         */
/*                                                             */
/*  Eingabe: n       -  Zeilen- bzw. Spaltenanzahl der         */
/*                      Matrix A                               */
/*           mdiag_p -  Adressenfeld wie oben beschrieben      */
/*           height  -  Feld der "Schornsteinhoehen wie        */
/*                      oben beschrieben                       */
/*           eps     -  Schranke fuer Abbruchkriterium (wenn   */
/*                      ein Hauptdiagonalelement um das eps-   */
/*                      fach kleiner oder gleich wird als das  */
/*                      kleinste vorher entstandene Haupt-     */
/*                      diagonalelement, wird "nicht positiv   */
/*                      definite Matrix" vermutet, und die     */
/*                      Rechnung wird abgebrochen)             */
/*                                                             */
/*  Ausgabe:            Rechtsdreiecksmatrix R, die auf die    */
/*                      gleiche Art wie die Matrix A           */
/*                      gespeichert ist (Eingabematrix wird    */
/*                      ueberschrieben)                        */
/*                                                             */
/*  Return-Wert:        Indikator, der den Erfolg der          */
/*                      Berechnung anzeigt:                    */
/*                      = 1  -->  Matrix A ist positiv         */
/*                                definit, Berechnung von      */
/*                                R wurde ausgefuehrt          */
/*                      = 0  -->  Matrix A ist nicht posi-     */
/*                                tiv definit, Berechnung      */
/*                                wurde abgebrochen            */
/*                                                             */
/*  Autor: J. Dankert                                          */
/* *********************************************************** */

#include    <math.h>
#include    "femset_f.h"

int skycho_f (long n , double *mdiag_p[] , long height[] , double eps)
{
   long   i , j , janf , kmax ;
   double *idiag_p , *jdiag_p , *itop_p , *iact_p , *kend_p , *i_p , *k_p ;
   double s , smin = 0. ;
   
   fillin_f (1 , 0L , 0L) ;
      
   for (i = 0 ; i < n ; i++)            /* ... ueber alle Spalten            */
   {
     idiag_p = mdiag_p[i] ;             /* ... auf aktuelles Diagonalelement */ 
     itop_p  = idiag_p + height[i] ;    /* ... auf Spitze des aktuellen      */
                                        /*     Schornsteins                  */
     janf    = i - height[i] ;     /* ... erste abzuhandelnde Parallelspalte */
     jdiag_p = mdiag_p[janf++] ;   /* ... auf deren Diaginalelement          */
     
     if (height[i] > 0)
         *itop_p /= *jdiag_p ;     /* ... nur Division durch Diagonalelement */
     iact_p   = itop_p - 1 ;       /* der am weitesten entfernten relevanten */
     kmax     = 1 ;                /* Spalte                                 */
     
     for (j = janf ; j < i ; j++)  /* ... ueber alle weiteren Spalten links  */
     {                             /* von der i-ten Spalte                   */
       jdiag_p = mdiag_p[j] ;      /* ... auf Diagonalelement der Spalte j   */
       kend_p  = jdiag_p +         /* ... auf Spitze der j-Spalte, eventuell */
              ((height[j] > kmax) ? kmax : height[j]) ; /* "nicht ganz oben" */

       i_p     = iact_p + 1 ;
       s       = *iact_p ;
       
       for (k_p = jdiag_p + 1 ; k_p <= kend_p ; k_p++ , i_p++) 
       { 
          s -= *k_p * *i_p ;
       }   
            
       *iact_p = s / *jdiag_p ;
       
       iact_p-- ;
       kmax++   ;                     
     }
     
     /* Sonderbehandlung fuer Hauptdiagonalelement:                          */
     
     s = *idiag_p ;
     
     for (k_p = itop_p ; k_p > idiag_p ; k_p--) s -= *k_p * *k_p ;

     if (i == 0)
     {
        if (s <= 0.) return 0 ;
        smin = s ;
     }   
     else
     {
        if (s <= smin * eps) return 0 ;
        if (s <  smin)       smin = s ;
     }
     
     *idiag_p = sqrt (s) ;

     fillin_f (52 , i , height[i]) ;
   } 
      
   fillin_f (1000 , 0L , 0L) ;

   return 1 ;
}   
