/* *********************************************************** */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                 */
/*                                                             */
/*  Vorwaerts- und Rueckwaertseinsetzen mit "Skyline"-Matrix   */
/*  ========================================================   */
/*                                                             */
/*  Es wird das Vorwaerts- und Rueckwaertseinsetzen zur        */
/*  Loesung eines linearen Gleichungssystems mit symmetri-     */
/*  scher positiv definiter Matrix A ausgefuehrt. Die          */
/*  Cholesky-Zerlegung der Matrix A muss bereits ausge-        */
/*  fuehrt sein (z. B. von der Funktion skycho_f).             */
/*                                                             */
/*  skyslv_f liefert die Loesung X des Gleichungssystems       */
/*                                                             */
/*                           T                                 */
/*                    A x = R  R x = b,                        */
/*                                                             */
/*  wenn die Rechtsdreiecksmatrix R und der Vektor             */
/*  B bereitgestellt werden.                                   */
/*                                                             */
/*  Von R sind nur die wesentlichen Elemente in n              */
/*  eindimensionalen double-Arrays gespeichert:                */
/*                                                             */
/*  Jedes double-Array enthaelt die Werte eines "vertikalen    */
/*  Schornsteins", beginnend mit dem Hauptdiagonalelement,     */
/*  dem die darueber liegenden Elemente folgen. Die            */
/*  Anfangsadressen dieser Arrays (Adressen der Speicher-      */
/*  mit den Hauptdiagonalelementen) werden in einem Pointer-   */
/*  Array mdiag_p erwartet. Die Anzahl der darueber liegenden  */
/*  Elemente (ohne Hauptdiagonalelement) muessen in einem      */
/*  Array height angeliefert werden (ein Beispiel fuer die     */
/*  Speicherung solcher Matrizen findet sich bei               */
/*  der Beschreibung der Funktion skycho_f).                   */
/*                                                             */
/*  Eingabe: n       -  Zeilen- bzw. Spaltenanzahl der         */
/*                      Matrix R                               */
/*           mdiag_p -  Adressenfeld wie oben beschrieben      */
/*           height  -  Feld der "Schornsteinhoehen wie        */
/*                      oben beschrieben                       */
/*           b       -  Vektor ("rechte Seite" des             */
/*                      Gleichungssystems) mit n Elementen     */
/*                                                             */
/*  Ausgabe: x_p     -  Vektor (Loesung des Gleichungssystems) */
/*                      mit n Elementen                        */
/*                                                             */
/*  Autor: J. Dankert                                          */
/* *********************************************************** */

#include    "femset_f.h"

void skyslv_f (long n , double *mdiag_p[] , 
                        long     height[] , double b[] , double x[])
{
   long   i , j , k ;
   double *idiag_p , *itop_p , *k_p ;
   double s ;
   
   fillin_f (1 , 0L , 0L) ;

   /* Vorwaerteinsetzen: */
   
   for (i = 0 ; i < n ; i++)            /* ... ueber alle Spalten            */
   {
     idiag_p = mdiag_p[i] ;             /* ... auf aktuelles Diagonalelement */ 
     itop_p  = idiag_p + height[i] ;    /* ... auf Spitze des aktuellen      */ 
     s       = b[i] ;                   /*     Schornsteins                  */
     
     for (k_p = idiag_p + 1 , j = i - 1 ; k_p <= itop_p ; k_p++ , j--) 
     {
        s -= *k_p * x[j] ; 
     }
     
     x[i] = s / *idiag_p ;
     fillin_f (60 , i + 1 , 0L) ;
   } 

   /* Rueckwaerteinsetzen: */
   
   for (i = n - 1 ; i >= 0 ; i--)
   {   
      s = x[i] ;
      
      for (j = i + 1 , k = 1 ; j < n ; j++ , k++)
      {
         if (k <= height[j]) s -= *(mdiag_p[j] + k) * x[j] ;
      }
      
      x[i] = s / *(mdiag_p[i]) ;
      fillin_f (70 , i + 1 , 0L) ;
   }

   fillin_f (1000 , 0L , 0L) ;
}   
