/* *********************************************************** */
/*  ###  Finite-Elemente-Baukasten FEMSET  ###                 */
/*                                                             */
/*  Vorwaerts- und Rueckwaertseinsetzen mit Bandmatrix         */
/*  ==================================================         */
/*                                                             */
/*  Es wird das Vorwaerts- und Rueckwaertseinsetzen zur        */
/*  Loesung eines linearen Gleichungssystems mit symmetri-     */
/*  scher positiv definiter Matrix A ausgefuehrt. Die          */
/*  Cholesky-Zerlegung der Matrix A muss bereits ausge-        */
/*  fuehrt sein (z. B. von der Funktion cholba_f).             */
/*                                                             */
/*  vorrue_f liefert die Loesung X des Gleichungssystems       */
/*                                                             */
/*                           T                                 */
/*                    A X = R  R X = B,                        */
/*                                                             */
/*  wenn die Rechtsdreiecksmatrix R und die Rechteckmatrix     */
/*  B bereitgestellt werden (Beschreibung des Loesungs-        */
/*  verfahrens z. B. in  "Dankert: Numerische Methoden der     */
/*  Mechanik", Springer-Verlag 1977).                          */
/*                                                             */
/*  Von R sind nur die wesentlichen Elemente zeilenweise       */
/*  in einem eindimensionalen Feld r_p gespeichert (pro Zeile  */
/*  ib Elemente, ib ist die Bandweite). Die letzten ib-1       */
/*  Zeilen muessen durch Null-Elemente aufjeweils ib           */
/*  Elemente ergaenzt werden (ein Beispiel fuer die            */
/*  Speicherung solcher Matrizen findet sich bei               */
/*  der Beschreibung des Unterprogramms cholba_f).             */
/*                                                             */
/*  Eingabe:   r_p -  Bandmatrix, wie sie als Ergebnis         */
/*                    vom Unterprogramm cholba_f abgelie-      */
/*                    fert wird (eindimensionales Feld         */
/*                    mit n*ib Elementen)                      */
/*             b_p -  Rechteckmatrix ("rechte Seite" des       */
/*                    Gleichungssystems), zweidimensionales    */
/*                    Feld mit n Zeilen und nb Spalten         */
/*                    (Matrix B)                               */
/*             n   -  Zeilen- bzw. Spaltenanzahl der           */
/*                    Matrix R, Zeilenanzahl der Matrizen      */
/*                    B und X                                  */
/*             ib  -  Bandweite der Matrix R                   */
/*             nb  -  Spaltenanzahl der Matrizen B und X       */
/*                    ("Anzahl der Vektoren der rechten        */
/*                    Seite")                                  */
/*                                                             */
/*  Ausgabe    x_p -  Rechteckmatrix (Loesung des Glei-        */
/*                    chungssystems), zweidimensionales        */
/*                    Feld mit n Zeilen und nb Spalten, auf    */
/*                    dem die Loesung X des Gleichungssystems  */
/*                    steht, x_p und b_p duerfen beim Aufruf   */
/*                    auf die gleiche Adresse zeigen,          */
/*                    in diesem Fall wird B von X ueber-       */
/*                    schrieben                                */
/*                                                             */
/*  Autor: J. Dankert                                          */
/* *********************************************************** */

#include    <math.h>
#include    "femset_f.h"

void vorrue_f (double *r_p , double *b_p , double *x_p ,
               long    n   , long    ib  , long    nb)
{
   double     s   ;
   long       ib1 , nnb , i , j , k , ii , kk , kkk , idiag , z ;
   
   fillin_f (1 , 0L , 0L) ;

   ib1 = ib -  1 ;
   nnb = n  * nb ;

   if (ib == 1) 
   {                                    /* Sonderfall "Diagonalmatrix"       */
     for (i = 0 ; i < nb ; i++)         /* Schleife ueber alle Spalten von B */
       for (j = 0 , ii = i ; j < n ; j++ , ii += nb)
       {
         *(x_p+ii) = *(b_p+ii) / (*(r_p+j) * *(r_p+j)) ;
         fillin_f (60 , j + 1 , 0L) ;
         fillin_f (70 , j + 1 , 0L) ;
       }  
   }
   else
   {
     for (i = 0 ; i < nb ; i++)         /* Schleife ueber alle Spalten von B */
     {
        *(x_p+i) = *(b_p+i) / *r_p ;        /* Beginn "Vorwaertseinsetzen"   */
        idiag    = ib ;
        
        z = 1 ;
        fillin_f (60 , 1 , 0L) ;
        
        for (j = nb+i ; j < nnb ; j += nb)  /* ... Elemente einer "B-Spalte" */
        {
          s = *(b_p+j) ;

          for (k = 1 , kk = j-nb , kkk = idiag-ib1 ;
               k < ib && kk >= 0 ; k++ , kk -= nb , kkk -= ib1)
            s -= *(x_p+kk) * *(r_p+kkk) ;

          *(x_p+j) = s / *(r_p+idiag) ;
          idiag   += ib               ;
          
          z++ ;
          fillin_f (60 , z , 0L) ;
        }

        idiag      = (n - 1) * ib     ;     /* Beginn "Rueckwaertseinsetzen" */
        ii         = (n - 1) * nb + i ;     /* Position in X "ganz unten"    */
        *(x_p+ii) /= *(r_p+idiag)     ;
        
        z = n ;
        fillin_f (70 , n , 0L) ;

        for (j = ii-nb ; j >= 0 ; j -= nb)  /* ... Elemente einer "B-Spalte" */
        {
          idiag -= ib       ;
          s      = *(x_p+j) ;

          for (k = 1 , kk = j+nb , kkk = idiag+1 ;
               k < ib && kk < nnb ; k++ , kk += nb , kkk++)
            s -= *(x_p+kk) * *(r_p+kkk) ;

          *(x_p+j) = s / *(r_p+idiag) ;
          
          z-- ;
          fillin_f (70 , z , 0L) ;
        }
      }
   }   

   fillin_f (1000 , 0L , 0L) ;
}
